"""Some helper functions for khard"""

from datetime import datetime
import pathlib
import random
import string
from typing import Any, Optional, Sequence, Union

from ruamel.yaml.scalarstring import LiteralScalarString
from .typing import list_to_string, PostAddress


YamlPostAddresses = dict[str, Union[list[dict[str, Any]], dict[str, Any]]]


def pretty_print(table: list[list[str]], justify: str = "L") -> str:
    """Converts a list of lists into a string formatted like a table
    with spaces separating fields and newlines separating rows"""
    # support for multiline columns
    line_break_table = []
    for row in table:
        # get line break count
        most_line_breaks_in_row = 0
        for col in row:
            if str(col).count("\n") > most_line_breaks_in_row:
                most_line_breaks_in_row = col.count("\n")
        # fill table rows
        for index in range(0, most_line_breaks_in_row+1):
            line_break_row = []
            for col in row:
                try:
                    line_break_row.append(str(col).split("\n")[index])
                except IndexError:
                    line_break_row.append("")
            line_break_table.append(line_break_row)
    # replace table variable
    table = line_break_table
    # get width for every column
    column_widths = [0] * len(table[0])
    offset = 3
    for row in table:
        for index, col in enumerate(row):
            width = len(str(col))
            if width > column_widths[index]:
                column_widths[index] = width
    table_row_list = []
    for row in table:
        single_row_list = []
        for col_index, col in enumerate(row):
            if justify == "R":  # justify right
                formated_column = str(col).rjust(column_widths[col_index] +
                                                 offset)
            elif justify == "L":  # justify left
                formated_column = str(col).ljust(column_widths[col_index] +
                                                 offset)
            elif justify == "C":  # justify center
                formated_column = str(col).center(column_widths[col_index] +
                                                  offset)
            single_row_list.append(formated_column)
        table_row_list.append(' '.join(single_row_list))
    return '\n'.join(table_row_list)


def get_random_uid() -> str:
    return ''.join([random.choice(string.ascii_lowercase + string.digits)
                    for _ in range(36)])


def yaml_clean(value: Union[str, Sequence, dict[str, Any], None]
               ) -> Union[Sequence, str, dict[str, Any], LiteralScalarString,
                          None]:
    """
    sanitize yaml values according to some simple principles:
      1. empty values are none, so ruamel does not print an empty list/str
      2. list with only one item become this item
      3. multiline strings use the YAML literal style:
         https://yaml.org/spec/1.2.2/#literal-style

    :param value: the value to be sanitized
    :returns: the sanitized value
    """
    # special case for empty values
    if not value:
        return None

    if isinstance(value, list):
        # special case for single item lists:
        if len(value) == 1 and isinstance(value[0], str):
            return value[0]
    elif isinstance(value, str):
        if "\n" in value:
            return LiteralScalarString(value)

    return value


def yaml_dicts(
        data: Optional[dict[str, Any]],
        defaults: Union[dict[str, Any], list[str], None] = None
    ) -> Optional[dict[str, Any]]:
    """
    format a dict according to template, if empty use specified defaults

    :param data: dict of contact data with keys as types and values as values.
    :param defaults: default dict to be used if data is empty.
    :returns: dict of types and values.
    """
    if not data:
        if isinstance(defaults, list):
            return {key: None for key in defaults}

        return defaults

    data_dict = {}
    for key, value_list in sorted(data.items(), key=lambda k: k[0].lower()):
        data_dict[key] = value_list[0] if len(value_list) == 1 else value_list

    return data_dict


def yaml_addresses(addresses: Optional[dict[str, list[PostAddress]]],
                   address_properties: list[str],
                   defaults: Optional[list[str]] = None
                   ) -> Optional[YamlPostAddresses]:
    """
    build a dict from an address, using a list of properties, an address has.

    :param addresses: dict of addresses with existing properties
    :param address_properties: list of properties that make up an address
    :param defaults: list of address types
    :returns: dict of address types with all properties
    """
    if not addresses:
        if not defaults:
            return None

        address_fields = {key: None for key in address_properties}
        return {address_type: address_fields for address_type in defaults}

    address_dict: YamlPostAddresses = {}
    for address_type, addresses_ in addresses.items():
        entry = [
            {key: yaml_clean(address.get(f"{key[0].lower()}{key[1:]}"))
                for key in address_properties}
            for address in addresses_
        ]
        if len(entry) == 1:
            address_dict[address_type] = entry[0]
        else:
            address_dict[address_type] = entry
    return address_dict


def yaml_anniversary(anniversary: Union[str, datetime, None],
                     version: str) -> Optional[str]:
    """
    format an anniversary according to its contents and the vCard version.

    :param anniversary: a string or a datetime object, that is the anniversary
    :param version: the vCard version to format for
    :returns: the formatted date string
    """
    if not anniversary:
        return None

    if isinstance(anniversary, datetime):
        if (version == "4.0" and anniversary.year == 1900
                             and anniversary.month != 0
                             and anniversary.day != 0
                             and anniversary.hour == 0
                             and anniversary.minute == 0
                             and anniversary.second == 0):
            return anniversary.strftime("--%m-%d")

        time_zone = anniversary.tzname()
        if ((time_zone and time_zone[3:])
                or anniversary.hour != 0
                or anniversary.minute != 0
                or anniversary.second != 0):
            return anniversary.isoformat()

        return anniversary.strftime("%F")

    # default case: anniversary is a string
    return anniversary


def convert_to_yaml(name: str, value: Union[None, str, list], indentation: int,
                    index_of_colon: int, show_multi_line_character: bool
                    ) -> list[str]:
    """converts a value list into yaml syntax

    :param name: name of object (example: phone)
    :param value: object contents
    :type value: str, list(str), list(list(str)), list(dict)
    :param indentation: indent all by number of spaces
    :param index_of_colon: use to position : at the name string (-1 for no
        space)
    :param show_multi_line_character: option to hide "|"
    :returns: yaml formatted string array of name, value pair
    """
    strings = []
    if isinstance(value, list):
        # special case for single item lists:
        if len(value) == 1 and isinstance(value[0], str):
            # value = ["string"] should not be converted to
            # name:
            #   - string
            # but to "name: string" instead
            value = value[0]
        elif len(value) == 1 and isinstance(value[0], list) \
                and len(value[0]) == 1 and isinstance(value[0][0], str):
            # same applies to value = [["string"]]
            value = value[0][0]
    if isinstance(value, str):
        strings.append("{}{}{}: {}".format(
            ' ' * indentation, name, ' ' * (index_of_colon-len(name)),
            indent_multiline_string(value, indentation+4,
                                    show_multi_line_character)))
    elif isinstance(value, list):
        strings.append("{}{}{}: ".format(
            ' ' * indentation, name, ' ' * (index_of_colon-len(name))))
        for outer in value:
            # special case for single item sublists
            if isinstance(outer, list) and len(outer) == 1 \
                    and isinstance(outer[0], str):
                # outer = ["string"] should not be converted to
                # -
                #   - string
                # but to "- string" instead
                outer = outer[0]
            if isinstance(outer, str):
                strings.append("{}- {}".format(
                    ' ' * (indentation+4), indent_multiline_string(
                        outer, indentation+8, show_multi_line_character)))
            elif isinstance(outer, list):
                strings.append("{}- ".format(' ' * (indentation+4)))
                for inner in outer:
                    if isinstance(inner, str):
                        strings.append("{}- {}".format(
                            ' ' * (indentation+8), indent_multiline_string(
                                inner, indentation+12,
                                show_multi_line_character)))
            elif isinstance(outer, dict):
                # ABLABEL'd lists
                for k in outer:
                    strings += convert_to_yaml(
                        "- " + k, outer[k], indentation+4, index_of_colon,
                        show_multi_line_character)
    return strings


def indent_multiline_string(input: Union[str, list], indentation: int,
                            show_multi_line_character: bool) -> str:
    # if input is a list, convert to string first
    if isinstance(input, list):
        input = list_to_string(input, "")
    # format multiline string
    if "\n" in input or ": " in input:
        lines = ["|"] if show_multi_line_character else [""]
        for line in input.split("\n"):
            lines.append("{}{}".format(' ' * indentation, line.strip()))
        return '\n'.join(lines)
    return input.strip()


def get_new_contact_template(
        supported_private_objects: Optional[list[str]] = None) -> str:
    formatted_private_objects = []
    if supported_private_objects:
        formatted_private_objects.append("")
        longest_key = max(supported_private_objects, key=len)
        for object in supported_private_objects:
            formatted_private_objects += convert_to_yaml(
                object, "", 12, len(longest_key)+1, True)
    template = pathlib.Path(__file__).parent.parent / 'data' / 'template.yaml'
    with template.open() as temp:
        return temp.read().format('\n'.join(formatted_private_objects))



"""Queries to match against contacts"""

import abc
from datetime import datetime
from functools import reduce
from operator import and_, or_
import re
from typing import cast, Any, Optional, Union

from . import contacts

# constants
FIELD_PHONE_NUMBERS = "phone_numbers"


class Query(metaclass=abc.ABCMeta):

    """A query to match against strings, lists of strings and Contacts"""

    @abc.abstractmethod
    def match(self, thing: Union[str, "contacts.Contact"]) -> bool:
        """Match the self query against the given thing"""

    @abc.abstractmethod
    def get_term(self) -> Optional[str]:
        """Extract the search terms from a query."""

    def __and__(self, other: "Query") -> "Query":
        """Combine two queries with AND"""
        if isinstance(self, NullQuery) or isinstance(other, NullQuery):
            return NullQuery()
        if isinstance(self, AnyQuery):
            return other
        if isinstance(other, AnyQuery):
            return self
        if isinstance(self, AndQuery) and isinstance(other, AndQuery):
            return AndQuery(*self._queries, *other._queries)
        if isinstance(self, AndQuery):
            return AndQuery(*self._queries, other)
        if isinstance(other, AndQuery):
            return AndQuery(self, *other._queries)
        return AndQuery(self, other)

    def __or__(self, other: "Query") -> "Query":
        """Combine two queries with OR"""
        if isinstance(self, AnyQuery) or isinstance(other, AnyQuery):
            return AnyQuery()
        if isinstance(self, NullQuery):
            return other
        if isinstance(other, NullQuery):
            return self
        if isinstance(self, OrQuery) and isinstance(other, OrQuery):
            return OrQuery(*self._queries, *other._queries)
        if isinstance(self, OrQuery):
            return OrQuery(*self._queries, other)
        if isinstance(other, OrQuery):
            return OrQuery(self, *other._queries)
        return OrQuery(self, other)

    def __eq__(self, other: object) -> bool:
        """A generic equality for all query types without parameters"""
        return isinstance(other, type(self))

    def __hash__(self) -> int:
        "A generic hashing implementation for all queries without parameters"
        return hash(type(self))


class NullQuery(Query):

    """The null-query, it matches nothing."""

    def match(self, thing: Union[str, "contacts.Contact"]) -> bool:
        return False

    def get_term(self) -> None:
        return None

    def __str__(self) -> str:
        return "NONE"


class AnyQuery(Query):

    """The match-anything-query, it always matches."""

    def match(self, thing: Union[str, "contacts.Contact"]) -> bool:
        return True

    def get_term(self) -> str:
        return ""

    def __hash__(self) -> int:
        return hash(NullQuery)

    def __str__(self) -> str:
        return "ALL"


class TermQuery(Query):

    """A query to match an object against a fixed string."""

    def __init__(self, term: str) -> None:
        self._term = term.lower()

    def match(self, thing: Union[str, "contacts.Contact"]) -> bool:
        if isinstance(thing, str):
            return self._term in thing.lower()
        return self._term in thing.pretty().lower()

    def get_term(self) -> str:
        return self._term

    def __eq__(self, other: object) -> bool:
        return isinstance(other, TermQuery) and self._term == other._term

    def __hash__(self) -> int:
        return hash((TermQuery, self._term))

    def __str__(self) -> str:
        return self._term


class FieldQuery(TermQuery):

    """A query to match against a certain field in a contact object."""

    def __init__(self, field: str, value: str) -> None:
        self._field = field
        super().__init__(value)

    def match(self, thing: Union[str, "contacts.Contact"]) -> bool:
        if isinstance(thing, str):
            return super().match(thing)
        if hasattr(thing, self._field):
            return self._match_union(getattr(thing, self._field))
        return False

    def _match_union(self, value: Union[str, datetime, list, dict[str, Any]]
                     ) -> bool:
        if isinstance(value, str):
            return self.match(value)
        if isinstance(value, list):
            return any(self._match_union(item) for item in value)
        if isinstance(value, dict):
            for key in value:
                if self.match(key) or self._match_union(value[key]):
                    return True
            return False
        if isinstance(value, datetime):
            # currently we only support ISO dates
            return value == datetime.strptime(self._term, "%Y-%m-%d")
        # this should actually be a type error
        return False

    def __eq__(self, other: object) -> bool:
        return isinstance(other, FieldQuery) and self._field == other._field \
            and super().__eq__(other)

    def __hash__(self) -> int:
        return hash((FieldQuery, self._field, self._term))

    def __str__(self) -> str:
        return '{}:{}'.format(self._field, self._term)


class AndQuery(Query):

    """A query to combine multiple queries with "and"."""

    def __init__(self, first: Query, second: Query, *queries: Query) -> None:
        self._queries = (first, second, *queries)

    def match(self, thing: Union[str, "contacts.Contact"]) -> bool:
        return all(q.match(thing) for q in self._queries)

    def get_term(self) -> Optional[str]:
        terms = [x.get_term() for x in self._queries]
        if None in terms:
            return None
        return "".join(cast(list[str], terms))

    def __eq__(self, other: object) -> bool:
        return isinstance(other, AndQuery) \
            and frozenset(self._queries) == frozenset(other._queries)

    def __hash__(self) -> int:
        return hash((AndQuery, frozenset(self._queries)))

    @staticmethod
    def reduce(queries: list[Query], start: Optional[Query] = None) -> Query:
        return reduce(and_, queries, start or AnyQuery())

    def __str__(self) -> str:
        return ' '.join(str(q) for q in self._queries)


class OrQuery(Query):

    """A query to combine multiple queries with "or"."""

    def __init__(self, first: Query, second: Query, *queries: Query) -> None:
        self._queries = (first, second, *queries)

    def match(self, thing: Union[str, "contacts.Contact"]) -> bool:
        return any(q.match(thing) for q in self._queries)

    def get_term(self) -> Optional[str]:
        terms = [x.get_term() for x in self._queries]
        if all(t is None for t in terms):
            return None
        return "".join(filter(None, terms))

    def __eq__(self, other: object) -> bool:
        return isinstance(other, OrQuery) \
            and frozenset(self._queries) == frozenset(other._queries)

    def __hash__(self) -> int:
        return hash((OrQuery, frozenset(self._queries)))

    @staticmethod
    def reduce(queries: list[Query], start: Optional[Query] = None) -> Query:
        return reduce(or_, queries, start or NullQuery())

    def __str__(self) -> str:
        return ' | '.join(str(q) for q in self._queries)


class NameQuery(TermQuery):

    """special query to match any kind of name field of a vCard"""

    def __init__(self, term: str) -> None:
        super().__init__(term)
        self._props_query = OrQuery(FieldQuery("formatted_name", term),
                                    FieldQuery("nicknames", term))

    def match(self, thing: Union[str, "contacts.Contact"]) -> bool:
        m = super().match
        if isinstance(thing, str):
            return m(thing)
        return (m(thing.get_first_name_last_name()) or
                m(thing.get_last_name_first_name()) or
                self._props_query.match(thing))

    def __eq__(self, other: object) -> bool:
        return isinstance(other, NameQuery) and self._term == other._term

    def __hash__(self) -> int:
        return hash((NameQuery, self._term))

    def __str__(self) -> str:
        return 'name:{}'.format(self._term)


class PhoneNumberQuery(FieldQuery):

    """A special query to match against phone numbers."""

    @staticmethod
    def _strip_phone_number(number: str) -> str:
        return re.sub("[^0-9+]", "", number)

    def __init__(self, value: str) -> None:
        super().__init__(FIELD_PHONE_NUMBERS, value)
        self._term_only_digits = self._strip_phone_number(value)

    def match(self, thing: Union[str, "contacts.Contact"]) -> bool:
        if isinstance(thing, str):
            return self._match_union(thing)
        else:
            return super().match(thing)

    def _match_union(self, value: Union[str, datetime, list, dict[str, Any]]
                     ) -> bool:
        if isinstance(value, str):
            if self._term in value.lower() \
                    or self._match_phone_number(self._strip_phone_number(value)):
                return True
        if isinstance(value, dict):
            for key in value:
                if self._term in str(key).lower():
                    return True
                if isinstance(value[key], str):
                    if self._match_phone_number(
                            self._strip_phone_number(value[key])):
                        return True
                if isinstance(value[key], list):
                    for number in value[key]:
                        if self._match_phone_number(
                                self._strip_phone_number(number)):
                            return True
            return False
        # this should actually be a type error
        return False

    def _match_phone_number(self, number: str) -> bool:
        if self._term_only_digits.startswith("+") and number.startswith("+"):
            # _term_only_digits: +49123456789
            # number: +49123456789
            return self._term_only_digits in number
        elif self._term_only_digits.startswith("+") and number.startswith("0"):
            # assume, that _term_only_digits contains a complete phone number
            # _term_only_digits: +49123456789
            # number: 0123456789
            return number[1:] in self._term_only_digits
        elif self._term_only_digits.startswith("0") and number.startswith("+"):
            # can't assume, that _term_only_digits contains a complete phone number
            # _term_only_digits: 0123456789
            # number: +49123456789
            if len(self._term_only_digits) >= 5:
                # don't strip the leading "0" if the search term is too short
                # otherwise you may get false positives
                # _term could contain the latter part of a phone number instead
                return self._term_only_digits[1:] in number
        # end of special cases
        if self._term_only_digits:
            return self._term_only_digits in number
        return False

    def __eq__(self, other: object) -> bool:
        return isinstance(other, PhoneNumberQuery) and self._term == other._term

    def __hash__(self) -> int:
        return hash((PhoneNumberQuery, self._term))

    def __str__(self) -> str:
        return 'phone numbers:{}'.format(self._term)


def parse(string: str) -> Union[TermQuery, FieldQuery]:
    """Parse a string into a query object

    The input string interpreted as a :py:class:`FieldQuery` if it starts with
    a valid property name of the
    :py:class:`~khard.contacts.Contact` class, followed by a colon
    and an arbitrary search term.  Otherwise it is interpreted as a
    :py:class:`TermQuery`.

    :param string: a string to parse into a query
    :returns: a FieldQuery if the string contains a valid field specifier, a
        TermQuery otherwise
    """
    if ":" in string:
        field, term = string.split(":", maxsplit=1)
        if field == "name":
            return NameQuery(term)
        if field == FIELD_PHONE_NUMBERS:
            return PhoneNumberQuery(term)
        if field == "kind":
            for kind in contacts.Contact._supported_kinds:
                if kind.startswith(term.lower()):
                    return FieldQuery(field, kind)
            return TermQuery(string)
        if field in contacts.Contact.get_properties():
            return FieldQuery(field, term)
    return TermQuery(string)



"""A simple class to load and manage the vcard files from disk."""

import abc
import binascii
from collections.abc import Mapping, Sequence
import glob
import logging
import os
from typing import Generator, Iterator, Optional, Union, overload

import vobject.base

from . import contacts
from .exceptions import AddressBookParseError
from .query import AnyQuery, Query


logger = logging.getLogger(__name__)


class AddressBook(metaclass=abc.ABCMeta):
    """The base class of all address book implementations."""

    def __init__(self, name: str) -> None:
        """:param name: the name to identify the address book"""
        self._loaded = False
        self.contacts: dict[str, "contacts.Contact"] = {}
        self._short_uids: Optional[dict[str,
                                        "contacts.Contact"]] = None
        self.name = name

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        return isinstance(other, type(self)) and self.name == other.name

    def __ne__(self, other: object) -> bool:
        return not self == other

    @staticmethod
    def _compare_uids(uid1: str, uid2: str) -> int:
        """Calculate the minimum length of initial substrings of uid1 and uid2
        for them to be different.

        :param uid1: first uid to compare
        :param uid2: second uid to compare
        :returns: the length of the shortest unequal initial substrings
        """
        return len(os.path.commonprefix((uid1, uid2)))

    def search(self, query: Query) -> Generator["contacts.Contact",
                                                None, None]:
        """Search this address book for contacts matching the query.

        The backend for this address book might be load()ed if needed.

        :param query: the query to search for
        :yields: all found contacts
        """
        logger.debug('address book %s, searching with %s', self.name, query)
        if not self._loaded:
            self.load(query)
        for contact in self.contacts.values():
            if query.match(contact):
                yield contact

    def get_short_uid_dict(self, query: Query = AnyQuery()) -> dict[
            str, "contacts.Contact"]:
        """Create a dictionary of shortened UIDs for all contacts.

        All arguments are only used if the address book is not yet initialized
        and will just be handed to self.load().

        :param query: see self.load()
        :returns: the contacts mapped by the shortest unique prefix of their
            UID
        """
        if self._short_uids is None:
            if not self._loaded:
                self.load(query)
            if not self.contacts:
                self._short_uids = {}
            elif len(self.contacts) == 1:
                self._short_uids = {uid[0:1]: contact
                                    for uid, contact in self.contacts.items()}
            else:
                self._short_uids = {}
                sorted_uids = sorted(self.contacts)
                # Prepare for the loop; the first and last items are handled
                # separately.
                item0, item1 = sorted_uids[:2]
                same1 = self._compare_uids(item0, item1)
                self._short_uids[item0[:same1 + 1]] = self.contacts[item0]
                for item_new in sorted_uids[2:]:
                    # shift the items and the common prefix length one further
                    item0, item1 = item1, item_new
                    same0, same1 = same1, self._compare_uids(item0, item1)
                    # compute the final prefix length for item1
                    same = max(same0, same1)
                    self._short_uids[item0[:same + 1]] = self.contacts[item0]
                # Save the last item.
                self._short_uids[item1[:same1 + 1]] = self.contacts[item1]
        return self._short_uids

    def get_short_uid(self, uid: str) -> str:
        """Get the shortened UID for the given UID.

        :param uid: the full UID to shorten
        :returns: the shortened uid or the empty string
        """
        if uid:
            short_uids = self.get_short_uid_dict()
            for length_of_uid in range(len(uid), 0, -1):
                if short_uids.get(uid[:length_of_uid]) is not None:
                    return uid[:length_of_uid]
        return ""

    @abc.abstractmethod
    def load(self, query: Query = AnyQuery()) -> None:
        """Load the vCards from the backing store.

        If a query is given loading is limited to entries which match the
        query.  If the query is None all entries will be loaded.

        :param query: the query to limit loading to matching entries
        :returns: the number of loaded contacts and the number of errors
        """


class VdirAddressBook(AddressBook):
    """An AddressBook implementation based on a vdir.

    This address book can load contacts from vcard files that reside in one
    directory on disk.
    """

    def __init__(self, name: str, path: str,
                 private_objects: Optional[list[str]] = None,
                 localize_dates: bool = True, skip: bool = False) -> None:
        """
        :param name: the name to identify the address book
        :param path: the path to the backing structure on disk
        :param private_objects: the names of private vCard extension fields to
            load
        :param localize_dates: whether to display dates in the local format
        :param skip: skip unparsable vCard files
        """
        self.path = os.path.expanduser(os.path.expandvars(path))
        if not os.path.isdir(self.path):
            raise NotADirectoryError("The path {} to the address book {} is "
                                     "not a directory".format(path, name))
        self._private_objects = private_objects or []
        self._localize_dates = localize_dates
        self._skip = skip
        super().__init__(name)

    def load(self, query: Query = AnyQuery(),
             search_in_source_files: bool = False) -> None:
        """Load all vcard files in this address book from disk.

        If a search string is given only files which contents match that will
        be loaded.

        :param query: query to limit the vcards that should be parsed
        :param search_in_source_files: apply search regexp directly on the .vcf
            files to speed up parsing (less accurate)
        :throws: AddressBookParseError
        """
        if self._loaded:
            return
        logger.debug('Loading Vdir %s with query %s', self.name, query)
        errors = 0
        for filename in glob.glob(os.path.join(self.path, "*.vcf")):
            try:
                card = contacts.Contact.from_file(
                    self, filename,
                    query if search_in_source_files else AnyQuery(),
                    self._private_objects, self._localize_dates)
                if card is None:
                    continue
            except (OSError, vobject.base.ParseError, binascii.Error) as err:
                verb = "open" if isinstance(err, OSError) else "parse"
                logger.error("Error: Could not %s file %s\n%s", verb, filename,
                             err)
                if self._skip:
                    errors += 1
                else:
                    raise AddressBookParseError(filename, self.name, err)
            else:
                uid = card.uid
                if not uid:
                    logger.warning("Card %s from address book %s has no UID "
                                   "and will not be available.", card,
                                   self.name)
                elif uid in self.contacts:
                    logger.warning(
                        "Card %s and %s from address book %s have the same "
                        "UID. The former will not be available.", card,
                        self.contacts[uid], self.name)
                else:
                    self.contacts[uid] = card
        self._loaded = True
        if errors:
            logger.warning(
                "%d of %d vCard files of address book %s could not be parsed.",
                errors, len(self.contacts) + errors, self)
        logger.debug('Loaded %s contacts from address book %s.',
                     len(self.contacts), self.name)


class AddressBookCollection(AddressBook, Mapping, Sequence):
    """A collection of several address books.

    This represents a temporary merge of the contact collections provided by
    the underlying address books.  On load, all contacts from all
    subaddressbooks are copied into a dict in this address book.  This allows
    this class to use all other methods from the parent AddressBook class.
    """

    def __init__(self, name: str, abooks: list[VdirAddressBook]) -> None:
        """
        :param name: the name to identify the address book
        :param abooks: a list of address books to combine in this collection
        """
        super().__init__(name)
        self._abooks = {ab.name: ab for ab in abooks}

    def load(self, query: Query = AnyQuery()) -> None:
        """Load the wrapped address books with the given parameters

        All parameters will be handed to VdirAddressBook.load.

        :param query: a query to limit the vcards that should be parsed
        :throws: AddressBookParseError
        """
        if self._loaded:
            return
        logger.debug('Loading collection %s with query %s', self.name, query)
        for abook in self._abooks.values():
            abook.load(query)
            for uid in abook.contacts:
                if uid in self.contacts:
                    logger.warning(
                        "Card %s from address book %s will not be available "
                        "because there is already another card with the same "
                        "UID: %s", abook.contacts[uid], abook, uid)
                else:
                    self.contacts[uid] = abook.contacts[uid]
        self._loaded = True
        logger.debug('Loaded %s contacts from address book %s.',
                     len(self.contacts), self.name)

    @overload
    def __getitem__(self, key: Union[int, str]) -> VdirAddressBook: ...
    @overload
    def __getitem__(self, key: slice) -> list[VdirAddressBook]: ...
    def __getitem__(self, key: Union[int, str, slice]
                    ) -> Union[VdirAddressBook, list[VdirAddressBook]]:
        """Get one or more of the backing address books by name or index

        :param key: the name of the address book to get or its index
        :returns: the matching address book(s)
        :throws: KeyError
        """
        if isinstance(key, str):
            return self._abooks[key]
        return list(self._abooks.values())[key]

    def __iter__(self) -> Iterator[VdirAddressBook]:
        """:return: an iterator over the underlying address books"""
        return iter(self._abooks.values())

    def __len__(self) -> int:
        return len(self._abooks)



"""Formatting and sorting of contacts"""

from .contacts import Contact


class Formatter:

    """A formatter for Contact.

    It receives some settings on initialisation which influence the formatting
    of the contact.
    """

    FIRST = "first_name"
    LAST = "last_name"
    FORMAT = "formatted_name"

    def __init__(self, display: str, preferred_email: list[str],
                 preferred_phone: list[str], show_nicknames: bool,
                 parsable: bool) -> None:
        self._display = display
        self._preferred_email = preferred_email
        self._preferred_phone = preferred_phone
        self._show_nicknames = show_nicknames
        self._parsable = parsable

    @staticmethod
    def format_labeled_field(field: dict[str, list[str]], preferred: list[str]
                             ) -> str:
        """Format a labeled field from a vCard for display, the first entry
        under the preferred label will be returned

        :param field: the labeled field, this must not be empty!
        :param preferred: the order of preferred labels
        :returns: the formatted field entry
        """
        # filter out preferred type if set in config file
        found = []
        for pref in preferred:
            for key in field:
                if pref.lower() in key.lower():
                    found.append(key)
            if found:
                break
        keys = found or [k for k in field if "pref" in k.lower()] \
            or field.keys()
        # get first key in alphabetical order
        first_key = sorted(keys, key=lambda k: k.lower())[0]
        return "{}: {}".format(first_key, sorted(field.get(first_key, []))[0])

    def get_special_field(self, vcard: Contact, field: str) -> str:
        """Returns certain fields with specific formatting options
            (for support of some list command options)."""
        if field == 'name':
            if self._display == self.FIRST:
                name = vcard.get_first_name_last_name()
            elif self._display == self.FORMAT:
                name = vcard.formatted_name
            else:
                name = vcard.get_last_name_first_name()
            if vcard.nicknames and self._show_nicknames and not self._parsable:
                return "{} (Nickname: {})".format(name, vcard.nicknames[0])
            return name
        if field == 'phone':
            if vcard.phone_numbers:
                return self.format_labeled_field(vcard.phone_numbers,
                                                 self._preferred_phone)
        if field == 'email':
            if vcard.emails:
                return self.format_labeled_field(vcard.emails,
                                                 self._preferred_email)
        if field == 'kind':
            return vcard.kind
        return ""

    @staticmethod
    def get_nested_field(vcard: Contact, field: str) -> str:
        """Returns the value of a nested field from a string

        get_nested_field(vcard,'emails.home.1') is equivalent to
        vcard.emails['home'][1].

        :param vcard: the contact from which to get the field
        :param field: a field specification
        :returns: the nested field, or the empty string if it didn't exist
        """
        attr_name = field.split('.')[0]
        val = ''
        if hasattr(vcard, attr_name):
            val = getattr(vcard, attr_name)
            # Loop through separate parts, changing val to be the head element.
            for partial in field.split('.')[1:]:
                if isinstance(val, dict) and partial in val:
                    val = val[partial]
                elif partial.isdigit() and isinstance(val, list) \
                        and len(val) > int(partial):
                    val = val[int(partial)]
                # TODO: Completely support case insensitive indexing
                elif isinstance(val, dict) and partial.upper() in val:
                    val = val[partial.upper()]
                else:
                    val = ''
        # Convert None and other falsy values to the empty string
        return val or ''



"""Helper code for type annotations and runtime type conversion."""

from datetime import datetime
from typing import Union, overload


# some type aliases
Date = Union[str, datetime]
StrList = Union[str, list[str]]
PostAddress = dict[str, str]


@overload
def convert_to_vcard(name: str, value: StrList, constraint: type[str]) -> str: ...
@overload
def convert_to_vcard(name: str, value: StrList, constraint: type[list]) -> list[str]: ...
@overload
def convert_to_vcard(name: str, value: StrList, constraint: None) -> StrList: ...
def convert_to_vcard(name: str, value: StrList, constraint: Union[None, type[str], type[list]]) -> StrList:
    """converts user input into vCard compatible data structures

    :param name: object name, only required for error messages
    :param value: user input
    :param constraint: set the accepted return type for vCard attribute
    :returns: cleaned user input, ready for vCard or a ValueError
    """
    if isinstance(value, str):
        if constraint is list:
            return [value.strip()]
        return value.strip()
    if isinstance(value, list):
        if constraint is str:
            raise ValueError(f"{name} must contain a string.")
        if not all(isinstance(entry, str) for entry in value):
            raise ValueError(f"{name} must not contain a nested list")
        # filter out empty list items and strip leading and trailing space
        return [x.strip() for x in value if x.strip()]
    if constraint is str:
        raise ValueError(f"{name} must be a string.")
    if constraint is list:
        raise ValueError(f"{name} must be a list with strings.")
    raise ValueError(f"{name} must be a string or a list with strings.")


def list_to_string(input: Union[str, list], delimiter: str) -> str:
    """converts list to string recursively so that nested lists are supported

    :param input: a list of strings and lists of strings (and so on recursive)
    :param delimiter: the delimiter to use when joining the items
    :returns: the recursively joined list
    """
    if isinstance(input, list):
        return delimiter.join(
            list_to_string(item, delimiter) for item in input)
    return input


def string_to_list(input: Union[str, list[str]], delimiter: str) -> list[str]:
    if isinstance(input, list):
        return input
    return [x.strip() for x in input.split(delimiter)]


def string_to_date(string: str) -> datetime:
    """Convert a date string into a date object.

    :param string: the date string to parse
    :returns: the parsed datetime object
    """
    # try date formats --mmdd, --mm-dd, yyyymmdd, yyyy-mm-dd and datetime
    # formats yyyymmddThhmmss, yyyy-mm-ddThh:mm:ss, yyyymmddThhmmssZ,
    # yyyy-mm-ddThh:mm:ssZ.
    for fmt in ("--%m%d", "--%m-%d", "%Y%m%d", "%Y-%m-%d", "%Y%m%dT%H%M%S",
                "%Y-%m-%dT%H:%M:%S", "%Y%m%dT%H%M%SZ", "%Y-%m-%dT%H:%M:%SZ"):
        try:
            return datetime.strptime(string, fmt)
        except ValueError:
            continue  # with the next format
    # try datetime formats yyyymmddThhmmsstz and yyyy-mm-ddThh:mm:sstz where tz
    # may look like -06:00.
    for fmt in ("%Y%m%dT%H%M%S%z", "%Y-%m-%dT%H:%M:%S%z"):
        try:
            return datetime.strptime(''.join(string.rsplit(":", 1)), fmt)
        except ValueError:
            continue  # with the next format
    raise ValueError
